#include <stdio.h>
#include <conio.h>
#include <windows.h>
#include <winuser.h>
#include <winsock.h>

#define MAX 4096

SOCKET call_socket(const char *hostname, unsigned short portnum)
{
  struct sockaddr_in sa;
  struct hostent *hp;
  SOCKET s;
  unsigned long addr;

  memset(&sa, 0, sizeof(sa));
  addr = inet_addr(hostname);
  if (addr != INADDR_NONE)
  {
    sa.sin_addr.s_addr = addr;
    sa.sin_family = AF_INET;
  }
  else
  {
    hp = gethostbyname(hostname);
    if (hp == NULL)
      return INVALID_SOCKET;
    memcpy((char *)&sa.sin_addr, hp->h_addr, hp->h_length);
    sa.sin_family = hp->h_addrtype;
  }
  sa.sin_port = htons((u_short)portnum);
  s = socket(sa.sin_family, SOCK_STREAM, 0);
  if (s == INVALID_SOCKET)
    return INVALID_SOCKET;
  
  if (connect(s, (struct sockaddr *)&sa, sizeof(sa)) == SOCKET_ERROR)
  {
    closesocket(s);
    return INVALID_SOCKET;
  }
  return s;
}

short send_data(SOCKET s, char *buf, short len)
{
  short bytes,
        num;

  bytes = 0;
  num = 0;
  while (bytes < len)
  {
    num = send(s, buf, len - bytes, 0);
    if (num > 0)
    {
      bytes += num;
      buf += num;
    }
    else 
    {
      if (num == SOCKET_ERROR)
        return SOCKET_ERROR;
    }
  }
  
  return bytes;
}

void send_line(SOCKET s, char *str)
{
  send_data(s, str, strlen(str));
}

short read_line(SOCKET s, char *buf, short max)
{
  short n;

  do
  {
    n = recv(s, buf, max, MSG_PEEK);
    if (n == SOCKET_ERROR)
      return SOCKET_ERROR;
  } while (strchr(buf, 10) == NULL);
  n = strchr(buf, 10) - buf;
  n = recv(s, buf, n + 1, 0);
  if (n == SOCKET_ERROR)
    return SOCKET_ERROR;
  buf[--n] = 0;
  
  return n;
}

short hidden_line(char *buf, short max)
{
  char c;
  short n;

  n = 0;
  do
  {
    c = getch();
    if (c == 0)
      c = getch();
    buf[n++] = c;
  } while ((n < max) && (c != 13));
  buf[--n] = 10;
  buf[++n] = 0;
  getchar();

  return n;
}

void main(int argc, char *argv[])
{
  WSADATA wsadata;
  SOCKET sock;
  SOCKADDR_IN local_sin;
  char buf[MAX],
       cmd[MAX];
  
  if ((argc != 2) && (argc != 3))
  {
    printf("\nPOP3 -- by David Phillips <david@acz.org>\n");
    printf("\nUsage: POP3 mail_server [port]\n");
    return;
  }
  
  if (WSAStartup(MAKEWORD(1, 1), &wsadata) != 0)
  {
    printf("Error initializing Winsock!\n");
    return;
  }

  if (argc == 2)
    sock = call_socket(argv[1], 110);
  else
      sock = call_socket(argv[1], atoi(argv[2]));
  if (sock == INVALID_SOCKET)
  {
    printf("Failed to connect to host!\n");
    WSACleanup();
    return;
  }

  read_line(sock, buf, MAX);
  printf("%s\n", buf);
  printf("USER ");
  fgets(buf, 128, stdin);
  strcpy(cmd, "USER ");
  strcat(cmd, buf);
  send_line(sock, cmd);
  read_line(sock, buf, MAX);
  printf("%s\n", buf);
  if (strncmp(buf, "+OK", 3))
  {
    closesocket(sock);
    WSACleanup();
    return;
  }
    
  printf("PASS ");
  fflush(stdout);
  hidden_line(buf, MAX);
  strcpy(cmd, "PASS ");
  strcat(cmd, buf);
  send_line(sock, cmd);
  read_line(sock, buf, MAX);
  printf("%s\n", buf);
  if (strncmp(buf, "+OK", 3))
  {
    closesocket(sock);
    WSACleanup();
    return;
  }

  do
  {
    printf("pop3> ");
    fgets(cmd, 128, stdin);
    if ((cmd[0] == '?') || (!strnicmp(cmd, "help", 4)))
    {
      printf("Commands:\n");
      printf("\tDELE\tmsg\t\tmark message for deletion\n");
      printf("\tLIST\t[msg]\t\tdisplay scan listing\n");
      printf("\tNOOP\t\t\tdo nothing\n");
      printf("\tQUIT\t\t\tterminate session and exit\n");
      printf("\tRETR\tmsg\t\tretrieve message\n");
      printf("\tRSET\t\t\treset any marked messages\n");
      printf("\tSTAT\t\t\tdisplay drop listing\n");
      printf("\tTOP\tmsg n\t\tdisplay top n lines of message\n");
      printf("\tUIDL\t[msg]\t\tdisplay unique-id listing\n");
      continue;
    }
    send_line(sock, cmd);
    read_line(sock, buf, MAX);
    if ((strnicmp(buf, "+OK", 3) == 0) &&
        ((((strnicmp(cmd, "LIST", 4) == 0) || 
        (strnicmp(cmd, "UIDL", 4) == 0)) &&
        (strlen(cmd) == 5)) ||
        ((strnicmp(cmd, "TOP", 3) == 0) ||
        (strnicmp(cmd, "RETR", 4) == 0))))
    {
      for(;;)
      {
        if (buf[0] == '.')
        {
          if ((buf[1] == 13) && (buf[2] == 0))
            break;
          printf("%s\n", &buf[1]);
        }
        else
            printf("%s\n", buf);
        read_line(sock, buf, MAX);
      }
    }
    printf("%s\n", buf);
  } while (strnicmp(cmd, "QUIT", 4));
  
  closesocket(sock);
  WSACleanup();
}
